################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from functools import singledispatch
from typing import Iterable, Mapping, Tuple, Union

import numpy as np
import pyquil
import sympy
from orquestra.quantum.circuits import _builtin_gates, _circuit, _gates
from orquestra.quantum.circuits.symbolic.sympy_expressions import (
    SYMPY_DIALECT,
    expression_from_sympy,
)
from orquestra.quantum.circuits.symbolic.translations import translate_expression

from ._expressions import QUIL_DIALECT, expression_from_pyquil


def _n_qubits_by_ops(ops: Iterable[_gates.GateOperation]):
    try:
        return max(qubit_i for op in ops for qubit_i in op.qubit_indices) + 1
    except ValueError:
        return 0


def _import_expression(pyquil_expr):
    return translate_expression(expression_from_pyquil(pyquil_expr), SYMPY_DIALECT)


def _export_expression(expr: sympy.Expr):
    return translate_expression(expression_from_sympy(expr), QUIL_DIALECT)


def _import_matrix(pyquil_matrix: np.ndarray):
    return sympy.Matrix(
        [
            [_import_expression(element) for element in row]
            for row in pyquil_matrix.tolist()
        ]
    )


def _export_matrix(matrix: sympy.Matrix):
    return [[_export_expression(element) for element in row] for row in matrix.tolist()]


def _import_gate_def(gate_def: pyquil.quilbase.DefGate):
    pyquil_params = tuple(map(_import_expression, gate_def.parameters or []))

    return _gates.CustomGateDefinition(
        gate_name=gate_def.name,
        matrix=_import_matrix(gate_def.matrix),
        params_ordering=pyquil_params,
    )


def import_from_pyquil(program: pyquil.Program):
    custom_names = [gate_def.name for gate_def in program.defined_gates]
    custom_defs = {
        gate_def.name: _import_gate_def(gate_def) for gate_def in program.defined_gates
    }
    if len(custom_names) != len(custom_defs):
        raise ValueError(
            "Can't import circuits with non-unique gate definition names to orquestra: "
            f"{custom_names}"
        )

    ops = [
        _import_gate(instr, custom_defs)
        for instr in program.instructions
        if isinstance(instr, pyquil.gates.Gate)
    ]
    return _circuit.Circuit(ops, _n_qubits_by_ops(ops))


def _import_gate(
    instruction: pyquil.gates.Gate,
    custom_gate_defs: Mapping[str, _gates.CustomGateDefinition],
) -> _gates.GateOperation:
    try:
        return _import_gate_via_name(instruction)
    except ValueError:
        pass

    try:
        return _import_custom_gate(instruction, custom_gate_defs)
    except ValueError:
        pass

    raise NotImplementedError(
        f"Importing instruction {instruction} from PyQuil is unsupported."
    )


def _ensure_is_gate(
    gate_ref: _builtin_gates.GateRef, params: Tuple[_gates.Parameter, ...]
) -> _gates.Gate:
    if isinstance(gate_ref, _gates.Gate):
        return gate_ref
    else:
        return gate_ref(*params)


def _import_gate_via_name(gate: pyquil.gates.Gate) -> _gates.GateOperation:
    try:
        zq_gate_ref = _builtin_gates.builtin_gate_by_name(gate.name)
    except KeyError:
        raise ValueError(f"Can't import {gate} as a built-in gate")

    zq_params = tuple(map(_import_expression, gate.params))
    zq_gate = _ensure_is_gate(zq_gate_ref, zq_params)

    for modifier in gate.modifiers:
        if modifier == "DAGGER":
            zq_gate = zq_gate.dagger
        elif modifier == "CONTROLLED":
            zq_gate = zq_gate.controlled(1)
    all_qubits = _import_pyquil_qubits(gate.qubits)

    operation = zq_gate(*all_qubits)
    if not isinstance(operation, _gates.GateOperation):
        raise ValueError()
    return operation


def _import_custom_gate(instruction, custom_gate_defs):
    try:
        gate_def = custom_gate_defs[instruction.name]
    except KeyError:
        raise ValueError(
            f"Custom gate definition for {instruction} is missing from custom gate "
            f"defs {custom_gate_defs}"
        )

    zq_params = tuple(map(_import_expression, instruction.params))
    zq_qubits = _import_pyquil_qubits(instruction.qubits)
    return gate_def(*zq_params)(*zq_qubits)


def _import_pyquil_qubits(qubits):
    return tuple(qubit.index for qubit in qubits)


def _export_orquestra_gate_definition(gate_def: _gates.CustomGateDefinition):
    return pyquil.quilbase.DefGate(
        gate_def.gate_name,
        _export_matrix(gate_def.matrix),
        list(map(_export_expression, gate_def.params_ordering)),
    )


def _create_pyquil_custom_gate_definitions(
    custom_gate_defs: Iterable[_gates.CustomGateDefinition],
):
    return {
        gate_def.gate_name: _export_orquestra_gate_definition(gate_def)
        for gate_def in custom_gate_defs
    }


@singledispatch
def _unwrap_gate(gate: _gates.Gate):
    return gate


@_unwrap_gate.register(_gates.ControlledGate)
@_unwrap_gate.register(_gates.Dagger)
def _unwrap_controlled_gate(gate: Union[_gates.ControlledGate, _gates.Dagger]):
    return _unwrap_gate(gate.wrapped_gate)


def _gate_definition_from_matrix_factory_gate(
    gate: _gates.MatrixFactoryGate,
) -> _gates.CustomGateDefinition:
    symbols = tuple(sympy.Symbol(f"theta_{i}") for i in range(len(gate.params)))
    template_matrix = gate.matrix_factory(*symbols)
    return _gates.CustomGateDefinition(
        gate_name=gate.name,
        matrix=template_matrix,
        params_ordering=symbols,
    )


def _is_builtin_gate(gate):
    return hasattr(_builtin_gates, gate.name)


def _is_supported_by_pyquil(gate):
    return hasattr(pyquil.gates, gate.name)


def _unique_by(sequence, key):
    return {key(obj): obj for obj in sequence}.values()


def _collect_unsupported_builtin_gate_defs(gates: Iterable[_gates.Gate]):
    unwrapped_gates = _unique_by(map(_unwrap_gate, gates), key=lambda gate: gate.name)
    return [
        _gate_definition_from_matrix_factory_gate(gate)
        for gate in unwrapped_gates
        if _is_builtin_gate(gate) and not _is_supported_by_pyquil(gate)
    ]


def export_to_pyquil(circuit: _circuit.Circuit) -> pyquil.Program:
    var_declarations = map(_param_declaration, sorted(map(str, circuit.free_symbols)))
    custom_gate_definitions = [
        *circuit.collect_custom_gate_definitions(),
        *_collect_unsupported_builtin_gate_defs([op.gate for op in circuit.operations]),
    ]
    pyquil_gate_definitions = _create_pyquil_custom_gate_definitions(
        custom_gate_definitions
    )

    gate_instructions = [
        _export_gate(op.gate, op.qubit_indices, pyquil_gate_definitions)
        for op in circuit.operations
    ]
    program = pyquil.Program(
        *[*var_declarations, *pyquil_gate_definitions.values(), *gate_instructions]
    )
    return program


def _param_declaration(param_name: str):
    return pyquil.quil.Declare(param_name, "REAL")


@singledispatch
def _export_gate(gate: _gates.Gate, qubit_indices, pyquil_gate_definitions):
    try:
        return _export_gate_via_name(gate, qubit_indices)
    except ValueError:
        pass

    return _export_custom_gate(gate, qubit_indices, pyquil_gate_definitions)


def _export_custom_gate(gate: _gates.Gate, qubit_indices, pyquil_gate_definitions):
    if gate.name not in pyquil_gate_definitions:
        raise ValueError(
            f"Can't export {gate} as custom gate, custom gate definition is missing"
        )
    constructor = pyquil_gate_definitions[gate.name].get_constructor()
    pyquil_params = list(map(_export_expression, gate.params))

    if pyquil_params:
        constructor = constructor(*pyquil_params)

    return constructor(*qubit_indices)


@_export_gate.register
def _export_controlled_gate(
    gate: _gates.ControlledGate, qubit_indices, custom_gate_names
):
    wrapped_qubit_indices = qubit_indices[gate.num_control_qubits :]
    control_qubit_indices = qubit_indices[0 : gate.num_control_qubits]
    exported = _export_gate(gate.wrapped_gate, wrapped_qubit_indices, custom_gate_names)
    for index in reversed(control_qubit_indices):
        exported = exported.controlled(index)
    return exported


@_export_gate.register
def _export_dagger(gate: _gates.Dagger, qubit_indices, custom_gate_names):
    return _export_gate(gate.wrapped_gate, qubit_indices, custom_gate_names).dagger()


def _pyquil_gate_by_name(name):
    return getattr(pyquil.gates, name)


def _export_gate_via_name(gate: _gates.Gate, qubit_indices):
    try:
        pyquil_fn = _pyquil_gate_by_name(gate.name)
    except AttributeError:
        raise ValueError(f"Can't export {gate} to PyQuil as a built-in gate")

    pyquil_params = map(_export_expression, gate.params)
    return pyquil_fn(*pyquil_params, *qubit_indices)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import re
import subprocess

import pytest

from bin import get_next_version


@pytest.mark.parametrize(
    "version_str",
    [
        # right after release
        "1.2.3",
        # a few commits after the release
        "0.1.1.dev73+g178469c.d20211213",
        # no release tag yet
        "0.1.dev82+gf24acc4",
    ],
)
def test_parse_version_str(version_str):
    semver = get_next_version.parse_version_str(version_str)
    assert semver.version_str == version_str


@pytest.mark.parametrize(
    "version_str,bumped",
    [
        # right after release
        ("1.2.3", "1.3.0"),
        # a few commits after the release
        ("0.1.1.dev73+g178469c.d20211213", "0.2.0"),
        # no release tag yet
        ("0.1.dev82+gf24acc4", "0.2.0"),
    ],
)
def test_bumping_version(version_str, bumped):
    semver = get_next_version.parse_version_str(version_str)
    assert semver.bump_minor.version_str == bumped



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pyquil
import pytest
import sympy
from orquestra.quantum.circuits import _builtin_gates, _circuit, _gates

from orquestra.integrations.forest.conversions import (
    export_to_pyquil,
    import_from_pyquil,
)

SYMPY_GAMMA = sympy.Symbol("gamma")
QUIL_GAMMA = pyquil.quil.Parameter("gamma")

"""
Note: Those differently named Symbols/Parameters are needed due
to the manner the conversion logic abstracts the names of the Symbols

Example:
SYMPY_THETA = sympy.Symbol("theta_0")
QUIL_THETA = pyquil.quil.Parameter("theta")

SYMPY_THETA = sympy.Symbol("theta")
QUIL_THETA = pyquil.quil.Parameter("theta")
----------------------------------------------
Export fails, Import passes
"""
SYMPY_THETA_0 = sympy.Symbol("theta_0")
SYMPY_THETA_1 = sympy.Symbol("theta_1")
SYMPY_THETA_2 = sympy.Symbol("theta_2")
QUIL_THETA_0 = pyquil.quil.Parameter("theta_0")
QUIL_THETA_1 = pyquil.quil.Parameter("theta_1")
QUIL_THETA_2 = pyquil.quil.Parameter("theta_2")


SQRT_X_DEF = _gates.CustomGateDefinition(
    "SQRT-X",
    sympy.Matrix([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]]),
    tuple(),
)

CUSTOM_PARAMETRIC_DEF = _gates.CustomGateDefinition(
    "CUSTOM-PARAMETRIC",
    sympy.Matrix(
        [
            [sympy.cos(SYMPY_GAMMA), sympy.sin(SYMPY_GAMMA)],
            [-sympy.sin(SYMPY_GAMMA), sympy.cos(SYMPY_GAMMA)],
        ]
    ),
    (SYMPY_GAMMA,),
)


def pyquil_xx_definition():
    return pyquil.quil.DefGate(
        name="XX",
        matrix=[
            [
                pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0),
                0,
                0,
                -1j * pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0),
            ],
            [
                0,
                pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0),
                -1j * pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0),
                0,
            ],
            [
                0,
                -1j * pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0),
                pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0),
                0,
            ],
            [
                -1j * pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0),
                0,
                0,
                pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0),
            ],
        ],
        parameters=[QUIL_THETA_0],
    )


def pyquil_rh_definition():
    cos_term = pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0)
    sin_term = pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0)
    phase_factor = 1j * sin_term + cos_term

    return pyquil.quil.DefGate(
        name="RH",
        matrix=[
            [
                phase_factor
                * (-0.5j * pyquil.quilatom.quil_sqrt(2) * sin_term + cos_term),
                -0.5j * pyquil.quilatom.quil_sqrt(2) * phase_factor * sin_term,
            ],
            [
                -0.5j * pyquil.quilatom.quil_sqrt(2) * phase_factor * sin_term,
                phase_factor
                * (0.5j * pyquil.quilatom.quil_sqrt(2) * sin_term + cos_term),
            ],
        ],
        parameters=[QUIL_THETA_0],
    )


def pyquil_u3_definition():
    # Note: need to add an extra global phase to match to orquestra's definition
    cos_term = pyquil.quilatom.quil_cos(0.5 * QUIL_THETA_0)
    sin_term = pyquil.quilatom.quil_sin(0.5 * QUIL_THETA_0)

    exp_phi = pyquil.quilatom.quil_exp(1j * QUIL_THETA_1)
    exp_lambda = pyquil.quilatom.quil_exp(1j * QUIL_THETA_2)
    exp_both = pyquil.quilatom.quil_exp(1j * (QUIL_THETA_1 + QUIL_THETA_2))

    return pyquil.quil.DefGate(
        name="U3",
        matrix=[
            [
                cos_term,
                -1 * exp_lambda * sin_term,
            ],
            [
                exp_phi * sin_term,
                cos_term * exp_both,
            ],
        ],
        parameters=[QUIL_THETA_0, QUIL_THETA_1, QUIL_THETA_2],
    )


PYQUIL_XX = pyquil_xx_definition()
PYQUIL_RH = pyquil_rh_definition()
PYQUIL_U3 = pyquil_u3_definition()


EQUIVALENT_CIRCUITS = [
    (
        _circuit.Circuit([], 0),
        pyquil.Program([]),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.X(2),
                _builtin_gates.Y(0),
            ]
        ),
        pyquil.Program([pyquil.gates.X(2), pyquil.gates.Y(0)]),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.CNOT(3, 1),
            ]
        ),
        pyquil.Program(
            [
                pyquil.gates.CNOT(3, 1),
            ]
        ),
    ),
    (
        _circuit.Circuit([_builtin_gates.RX(np.pi)(1)]),
        pyquil.Program([pyquil.gates.RX(np.pi, 1)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.S(1)]),
        pyquil.Program([pyquil.gates.S(1)]),
    ),
    (
        _circuit.Circuit(
            [_builtin_gates.SWAP.controlled(1)(2, 0, 3)],
        ),
        pyquil.Program([pyquil.gates.SWAP(0, 3).controlled(2)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.Y.controlled(2)(3, 1, 2)]),
        pyquil.Program([pyquil.gates.Y(2).controlled(1).controlled(3)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.RX(0.5).dagger.controlled(2)(3, 1, 2)]),
        pyquil.Program([pyquil.gates.RX(0.5, 2).dagger().controlled(1).controlled(3)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.RX(0.5).controlled(2).dagger(3, 1, 2)]),
        pyquil.Program([pyquil.gates.RX(0.5, 2).dagger().controlled(1).controlled(3)]),
    ),
    (
        _circuit.Circuit(
            [SQRT_X_DEF()(3)],
        ),
        pyquil.Program([("SQRT-X", 3)]).defgate(
            "SQRT-X",
            np.array(
                [
                    [0.5 + 0.5j, 0.5 - 0.5j],
                    [0.5 - 0.5j, 0.5 + 0.5j],
                ]
            ),
        ),
    ),
    (
        _circuit.Circuit([_builtin_gates.XX(np.pi)(2)]),
        pyquil.Program([PYQUIL_XX, PYQUIL_XX.get_constructor()(np.pi)(2)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.RH(np.pi / 5)(3)]),
        pyquil.Program([PYQUIL_RH, PYQUIL_RH.get_constructor()(np.pi / 5)(3)]),
    ),
    (
        _circuit.Circuit([_builtin_gates.U3(np.pi / 2, np.pi / 4, 0)(3)]),
        pyquil.Program(
            [PYQUIL_U3, PYQUIL_U3.get_constructor()(np.pi / 2, np.pi / 4, 0)(3)]
        ),
    ),
    (
        _circuit.Circuit(
            [_builtin_gates.U3(np.pi / 2, np.pi / 4, 0).controlled(1)(2, 3)]
        ),
        pyquil.Program(
            [
                PYQUIL_U3,
                PYQUIL_U3.get_constructor()(np.pi / 2, np.pi / 4, 0)(3).controlled(2),
            ]
        ),
    ),
]


def _example_parametric_pyquil_program():
    gate_def = pyquil.quil.DefGate(
        "CUSTOM-PARAMETRIC",
        [
            [
                pyquil.quilatom.quil_cos(QUIL_GAMMA),
                pyquil.quilatom.quil_sin(QUIL_GAMMA),
            ],
            [
                -pyquil.quilatom.quil_sin(QUIL_GAMMA),
                pyquil.quilatom.quil_cos(QUIL_GAMMA),
            ],
        ],
        [QUIL_GAMMA],
    )
    gate_constructor = gate_def.get_constructor()

    return pyquil.Program(
        pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
        gate_def,
        gate_constructor(QUIL_THETA_0)(0),
    )


def _example_rh_pyquil_program():
    gate_def = PYQUIL_RH
    gate_constructor = gate_def.get_constructor()

    return pyquil.Program(
        pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
        gate_def,
        gate_constructor(QUIL_THETA_0)(0),
    )


def _example_xx_pyquil_program():
    gate_def = PYQUIL_XX
    gate_constructor = gate_def.get_constructor()

    return pyquil.Program(
        pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
        gate_def,
        gate_constructor(QUIL_THETA_0)(0),
    )


def _example_u3_pyquil_program():
    gate_def = PYQUIL_U3
    gate_constructor = gate_def.get_constructor()

    return pyquil.Program(
        pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
        pyquil.quil.Declare(QUIL_THETA_1.name, "REAL"),
        pyquil.quil.Declare(QUIL_THETA_2.name, "REAL"),
        gate_def,
        gate_constructor(QUIL_THETA_0, QUIL_THETA_1, QUIL_THETA_2)(0),
    )


EQUIVALENT_PARAMETRIZED_CIRCUITS = [
    (
        _circuit.Circuit(
            [
                _builtin_gates.RX(SYMPY_THETA_0)(1),
            ],
        ),
        pyquil.Program(
            [
                pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
                pyquil.gates.RX(QUIL_THETA_0, 1),
            ]
        ),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.RX(
                    sympy.Mul(SYMPY_THETA_0, SYMPY_THETA_1, evaluate=False)
                )(1),
            ],
        ),
        pyquil.Program(
            [
                pyquil.quil.Declare(QUIL_THETA_0.name, "REAL"),
                pyquil.quil.Declare(QUIL_THETA_1.name, "REAL"),
                pyquil.gates.RX(QUIL_THETA_0 * QUIL_THETA_1, 1),
            ]
        ),
    ),
    (
        _circuit.Circuit(
            [
                CUSTOM_PARAMETRIC_DEF(SYMPY_THETA_0)(0),
            ],
        ),
        _example_parametric_pyquil_program(),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.RH(SYMPY_THETA_0)(0),
            ],
        ),
        _example_rh_pyquil_program(),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.XX(SYMPY_THETA_0)(0),
            ],
        ),
        _example_xx_pyquil_program(),
    ),
    (
        _circuit.Circuit(
            [
                _builtin_gates.U3(SYMPY_THETA_0, SYMPY_THETA_1, SYMPY_THETA_2)(0),
            ],
        ),
        _example_u3_pyquil_program(),
    ),
]


class TestExportingToPyQuil:
    @pytest.mark.parametrize(
        "orquestra_circuit, pyquil_circuit",
        [*EQUIVALENT_CIRCUITS, *EQUIVALENT_PARAMETRIZED_CIRCUITS],
    )
    def test_exporting_circuit_gives_equivalent_circuit(
        self, orquestra_circuit, pyquil_circuit
    ):
        exported = export_to_pyquil(orquestra_circuit)
        assert exported == pyquil_circuit, (
            exported.out(),
            pyquil_circuit.out(),
        )


class TestImportingFromPyQuil:
    @pytest.mark.parametrize(
        "orquestra_circuit, pyquil_circuit",
        [*EQUIVALENT_CIRCUITS, *EQUIVALENT_PARAMETRIZED_CIRCUITS],
    )
    def test_importing_circuit_gives_equivalent_circuit(
        self, orquestra_circuit, pyquil_circuit
    ):
        imported = import_from_pyquil(pyquil_circuit)
        assert imported == orquestra_circuit



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for symbolic_expressions module."""
import pytest
import sympy
from orquestra.quantum.circuits.symbolic.sympy_expressions import (
    SYMPY_DIALECT,
    expression_from_sympy,
)
from orquestra.quantum.circuits.symbolic.translations import translate_expression
from pyquil import quil, quilatom

from orquestra.integrations.forest.conversions._expressions import (
    QUIL_DIALECT,
    expression_from_pyquil,
)


@pytest.mark.parametrize(
    "sympy_expression, quil_expression",
    [
        (sympy.Symbol("theta"), quil.Parameter("theta")),
        (
            sympy.Mul(sympy.Symbol("theta"), sympy.Symbol("gamma"), evaluate=False),
            quil.Parameter("theta") * quil.Parameter("gamma"),
        ),
        (sympy.cos(sympy.Symbol("theta")), quilatom.quil_cos(quil.Parameter("theta"))),
        (
            sympy.cos(2 * sympy.Symbol("theta")),
            quilatom.quil_cos(2 * quil.Parameter("theta")),
        ),
        (
            sympy.exp(sympy.Symbol("x") - sympy.Symbol("y")),
            quilatom.quil_exp(quil.Parameter("x") - quil.Parameter("y")),
        ),
        (
            sympy.Add(
                sympy.cos(sympy.Symbol("phi")),
                sympy.I * sympy.sin(sympy.Symbol("phi")),
                evaluate=False,
            ),
            quilatom.quil_cos(quil.Parameter("phi"))
            + 1j * quilatom.quil_sin(quil.Parameter("phi")),
        ),
        (
            sympy.Add(
                sympy.Symbol("x"),
                sympy.Mul(sympy.Symbol("y"), (2 + 3j), evaluate=False),
                evaluate=False,
            ),
            quil.Parameter("x") + quil.Parameter("y") * (2 + 3j),
        ),
        (
            sympy.cos(sympy.sin(sympy.Symbol("tau"))),
            quilatom.quil_cos(quilatom.quil_sin(quil.Parameter("tau"))),
        ),
        (
            sympy.Symbol("x") / sympy.Symbol("y"),
            quil.Parameter("x") / quil.Parameter("y"),
        ),
        (
            sympy.tan(sympy.Symbol("theta")),
            quilatom.quil_sin(quil.Parameter("theta"))
            / quilatom.quil_cos(quil.Parameter("theta")),
        ),
        (2 ** sympy.Symbol("x"), 2 ** quil.Parameter("x")),
        (
            sympy.Symbol("y") ** sympy.Symbol("x"),
            quil.Parameter("y") ** quil.Parameter("x"),
        ),
        (sympy.Symbol("x") ** 2, quil.Parameter("x") ** 2),
        (
            sympy.sqrt(sympy.Symbol("x") - sympy.Symbol("y")),
            quilatom.quil_sqrt(quil.Parameter("x") - quil.Parameter("y")),
        ),
        (
            -5 * sympy.Symbol("x") * sympy.Symbol("y"),
            -5 * quil.Parameter("x") * quil.Parameter("y"),
        ),
    ],
)
def test_translating_tree_from_sympy_to_quil_gives_expected_result(
    sympy_expression, quil_expression
):
    expression = expression_from_sympy(sympy_expression)
    assert translate_expression(expression, QUIL_DIALECT) == quil_expression


@pytest.mark.parametrize(
    "quil_expression, sympy_expression",
    [
        (quil.Parameter("theta"), sympy.Symbol("theta")),
        (
            quil.Parameter("theta") * quil.Parameter("gamma"),
            sympy.Symbol("theta") * sympy.Symbol("gamma"),
        ),
        (
            quilatom.quil_cos(quil.Parameter("theta")),
            sympy.cos(sympy.Symbol("theta")),
        ),
        (
            quilatom.quil_cos(2 * quil.Parameter("theta")),
            sympy.cos(2 * sympy.Symbol("theta")),
        ),
        (
            quilatom.quil_exp(quil.Parameter("x") - quil.Parameter("y")),
            sympy.exp(sympy.Symbol("x") - sympy.Symbol("y")),
        ),
        (
            quilatom.quil_cos(quil.Parameter("phi"))
            + 1j * quilatom.quil_sin(quil.Parameter("phi")),
            (sympy.cos(sympy.Symbol("phi")) + sympy.I * sympy.sin(sympy.Symbol("phi"))),
        ),
        (
            quil.Parameter("x") + quil.Parameter("y") * (2 + 3j),
            sympy.Symbol("x") + sympy.Symbol("y") * (2 + 3j),
        ),
        (
            quilatom.quil_cos(quilatom.quil_sin(quil.Parameter("tau"))),
            sympy.cos(sympy.sin(sympy.Symbol("tau"))),
        ),
        (
            quil.Parameter("x") / quil.Parameter("y"),
            sympy.Symbol("x") / sympy.Symbol("y"),
        ),
        (2 ** quil.Parameter("x"), 2 ** sympy.Symbol("x")),
        (
            quil.Parameter("y") ** quil.Parameter("x"),
            sympy.Symbol("y") ** sympy.Symbol("x"),
        ),
        (quil.Parameter("x") ** 2, sympy.Symbol("x") ** 2),
        (
            quilatom.quil_sqrt(quil.Parameter("x") - quil.Parameter("y")),
            sympy.sqrt(sympy.Symbol("x") - sympy.Symbol("y")),
        ),
    ],
)
def test_translating_tree_from_quil_to_sympy_gives_expected_result(
    quil_expression, sympy_expression
):
    expression = expression_from_pyquil(quil_expression)
    assert translate_expression(expression, SYMPY_DIALECT) - sympy_expression == 0



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for decomposing PyQuil expression into our native expressions."""
import numpy as np
import pytest
from orquestra.quantum.circuits.symbolic.expressions import FunctionCall, Symbol
from pyquil import quil, quilatom

from orquestra.integrations.forest.conversions._expressions import (
    expression_from_pyquil,
)


@pytest.mark.parametrize("number", [3, 4.0, 1j, 3.0 - 2j])
def test_native_numbers_are_preserved(number):
    assert expression_from_pyquil(number) == number


@pytest.mark.parametrize(
    "pyquil_parameter, expected_symbol",
    [
        (quil.Parameter("theta"), Symbol("theta")),
        (quil.Parameter("x"), Symbol("x")),
        (quil.Parameter("x_1"), Symbol("x_1")),
    ],
)
def test_quil_parameters_are_converted_to_instance_of_symbol_with_correct_name(
    pyquil_parameter, expected_symbol
):
    assert expression_from_pyquil(pyquil_parameter) == expected_symbol


@pytest.mark.parametrize(
    "pyquil_function_call, expected_function_call",
    [
        (quilatom.quil_cos(2), FunctionCall("cos", (2,))),
        (
            quilatom.quil_sin(quil.Parameter("theta")),
            FunctionCall("sin", (Symbol("theta"),)),
        ),
        (quilatom.quil_exp(quil.Parameter("x")), FunctionCall("exp", (Symbol("x"),))),
        (quilatom.quil_sqrt(np.pi), FunctionCall("sqrt", (np.pi,))),
    ],
)
def test_pyquil_function_calls_are_converted_to_equivalent_function_call(
    pyquil_function_call, expected_function_call
):
    assert expression_from_pyquil(pyquil_function_call) == expected_function_call


@pytest.mark.parametrize(
    "pyquil_expression, expected_function_call",
    [
        (
            quil.Parameter("x") + quil.Parameter("y"),
            FunctionCall("add", (Symbol("x"), Symbol("y"))),
        ),
        (
            quilatom.quil_cos(quil.Parameter("theta")) * 2,
            FunctionCall("mul", (FunctionCall("cos", (Symbol("theta"),)), 2)),
        ),
        (
            quilatom.quil_sqrt(quil.Parameter("phi")) / quil.Parameter("psi"),
            FunctionCall(
                "div", (FunctionCall("sqrt", (Symbol("phi"),)), Symbol("psi"))
            ),
        ),
        (
            quil.Parameter("a") - quil.Parameter("b"),
            FunctionCall("sub", (Symbol("a"), Symbol("b"))),
        ),
        (2 ** quil.Parameter("N"), FunctionCall("pow", (2, Symbol("N")))),
    ],
)
def test_pyquil_binary_expressions_are_converted_to_appropriate_function_call(
    pyquil_expression, expected_function_call
):
    assert expression_from_pyquil(pyquil_expression) == expected_function_call



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
############################################################################
#   Copyright 2017 Rigetti Computing, Inc.
#   Modified by Zapata Computing 2020 for the performance reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
############################################################################
"""
Translates Orquestra pauli representation objects to pyQuil objects and vice versa.
"""
from typing import Union

from orquestra.quantum.operators import PauliRepresentation, PauliSum, PauliTerm
from pyquil.paulis import PauliSum as PyquilPauliSum
from pyquil.paulis import PauliTerm as PyquilPauliTerm


def orq_to_pyquil(
    pauli_operator: PauliRepresentation,
) -> Union[PyquilPauliSum, PyquilPauliTerm]:
    """
    Convert an Orquestra PauliSum or PauliTerm to a pyQuil PauliSum or PauliTerm,
    respectively.

    Args:
        pauli_operator: Orquestra PauliSum or PauliTerm to convert to a pyquil PauliSum
            or PauliTerm

    Returns:
        PauliSum or PauliTerm representing the input pauli operator
    """
    if not isinstance(pauli_operator, (PauliSum, PauliTerm)):
        raise TypeError(
            "pauli_operator must be an Orquestra PauliSum or PauliTerm object"
        )

    if isinstance(pauli_operator, PauliTerm):
        return _orq_to_pyquil_term(pauli_operator)

    terms = [_orq_to_pyquil_term(term) for term in pauli_operator.terms]
    paulisum = PyquilPauliSum(terms)

    return paulisum.simplify()


def pyquil_to_orq(
    pyquil_pauli: Union[PyquilPauliTerm, PyquilPauliSum]
) -> PauliRepresentation:
    """
    Convert a pyQuil PauliSum or PauliTerm to an Orquestra PauliSum or PauliTerm,
        respectively.

    Args:
        pyquil_pauli: pyQuil PauliTerm or PauliSum to convert to an Orquestra PauliTerm
            or PauliSum

    Returns:
        Orquestra PauliSum or PauliTerm representing the pyQuil PauliTerm or PauliSum
    """

    if not isinstance(pyquil_pauli, (PyquilPauliSum, PyquilPauliTerm)):
        raise TypeError("pyquil_pauli must be a pyquil PauliSum or PauliTerm object")

    if isinstance(pyquil_pauli, PyquilPauliTerm):
        return _pyquil_to_orq_term(pyquil_pauli)

    transformed_term = PauliSum()
    # iterate through the PauliTerms of PauliSum
    for pauli_term in pyquil_pauli.terms:
        transformed_term += _pyquil_to_orq_term(pauli_term)

    return transformed_term


def _orq_to_pyquil_term(orq_term: PauliTerm) -> PyquilPauliTerm:
    base_term = PyquilPauliTerm("I", 0)
    for idx, op in orq_term.operations:
        product = base_term * PyquilPauliTerm(op, idx)
        assert isinstance(product, PyquilPauliTerm)
        base_term = product
    return orq_term.coefficient * base_term


def _pyquil_to_orq_term(pyquil_term: PyquilPauliTerm) -> PauliTerm:
    try:
        return PauliTerm(pyquil_term._ops, pyquil_term.coefficient)  # type: ignore
    except TypeError:
        raise ValueError(
            "All qubit indices of pyQuil pauli must be integers. "
            "Offending term: {}".format(pyquil_term)
        )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Utilities related to Quil based symbolic expressions."""
import operator
from functools import singledispatch
from numbers import Number

import pyquil
from orquestra.quantum.circuits.symbolic.expressions import (
    ExpressionDialect,
    FunctionCall,
    Symbol,
    reduction,
)
from pyquil import quilatom

QUIL_BINARY_EXPRESSION_NAMES = {
    quilatom.Add: "add",
    quilatom.Sub: "sub",
    quilatom.Mul: "mul",
    quilatom.Div: "div",
    quilatom.Pow: "pow",
}


@singledispatch
def expression_from_pyquil(expression):
    raise NotImplementedError(
        f"Expression {expression} of type {type(expression)} is currently not supported"
    )


@expression_from_pyquil.register
def identity(number: Number):
    return number


@expression_from_pyquil.register
def symbol_from_quil_parameter(parameter: pyquil.quil.Parameter):
    return Symbol(parameter.name)


@expression_from_pyquil.register
def function_call_from_pyquil_function(function: pyquil.quilatom.Function):
    return FunctionCall(
        function.name.lower(), (expression_from_pyquil(function.expression),)
    )


@expression_from_pyquil.register(quilatom.Add)
@expression_from_pyquil.register(quilatom.Sub)
@expression_from_pyquil.register(quilatom.Mul)
@expression_from_pyquil.register(quilatom.Div)
@expression_from_pyquil.register(quilatom.Pow)
def function_call_from_pyquil_binary_expression(expression):
    return FunctionCall(
        QUIL_BINARY_EXPRESSION_NAMES[type(expression)],
        (
            expression_from_pyquil(expression.op1),
            expression_from_pyquil(expression.op2),
        ),
    )


# Dialect defining conversion of intermediate expression tree to
# the expression based on quil functions/parameters.
# This is intended to be passed by a `dialect` argument of `translate_expression`.
QUIL_DIALECT = ExpressionDialect(
    symbol_factory=lambda symbol: pyquil.quil.Parameter(symbol.name),
    number_factory=lambda number: number,
    known_functions={
        "add": reduction(operator.add),
        "mul": reduction(operator.mul),
        "div": operator.truediv,
        "sub": operator.sub,
        "pow": operator.pow,
        "cos": quilatom.quil_cos,
        "sin": quilatom.quil_sin,
        "exp": quilatom.quil_exp,
        "sqrt": quilatom.quil_sqrt,
        "tan": lambda arg: quilatom.quil_sin(arg) / quilatom.quil_cos(arg),
    },
)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
############################################################################
#   Copyright 2017 Rigetti Computing, Inc.
#   Modified by Zapata Computing 2020.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
############################################################################

import numpy as np
import pytest
from orquestra.quantum.operators import PauliSum as OrqPauliSum
from orquestra.quantum.operators import PauliTerm as OrqPauliTerm
from pyquil.paulis import PauliSum as PyquilPauliSum
from pyquil.paulis import PauliTerm as PyquilPauliTerm
from pyquil.quilatom import QubitPlaceholder

from orquestra.integrations.forest.conversions import orq_to_pyquil, pyquil_to_orq


def test_translation_type_enforcement():
    """
    Make sure type check works
    """
    sample_str = "Z0*Z1"
    sample_int = 1
    pyquil_term = PyquilPauliTerm("X", 0)
    pyquil_sum = PyquilPauliSum([pyquil_term])

    with pytest.raises(TypeError):
        orq_to_pyquil(sample_str)
    with pytest.raises(TypeError):
        orq_to_pyquil(sample_int)
    with pytest.raises(TypeError):
        orq_to_pyquil(pyquil_term)
    with pytest.raises(TypeError):
        orq_to_pyquil(pyquil_sum)

    # don't accept anything other than pyquil PauliSum or PauliTerm
    orq_term = OrqPauliTerm("X0*Y5")
    orq_sum = OrqPauliSum("0.5*X0*Z1*X2 + 0.5*Y0*Z1*Y2")
    with pytest.raises(TypeError):
        pyquil_to_orq(sample_str)
    with pytest.raises(TypeError):
        pyquil_to_orq(sample_int)
    with pytest.raises(TypeError):
        pyquil_to_orq(orq_term)
    with pytest.raises(TypeError):
        pyquil_to_orq(orq_sum)


def test_orq_paulisum_to_pyquil():
    """
    Conversion of PauliSum to pyquil; accuracy test
    """
    pauli_term = OrqPauliSum("0.5*X0*Z1*X2 + 0.5*Y0*Z1*Y2")

    forest_term = orq_to_pyquil(pauli_term)
    ground_truth = (
        PyquilPauliTerm("X", 0) * PyquilPauliTerm("Z", 1) * PyquilPauliTerm("X", 2)
    )
    ground_truth += (
        PyquilPauliTerm("Y", 0) * PyquilPauliTerm("Z", 1) * PyquilPauliTerm("Y", 2)
    )
    ground_truth *= 0.5

    assert ground_truth == forest_term
    assert isinstance(forest_term, PyquilPauliSum)


def test_orq_pauliterm_to_pyquil():
    """
    Conversion of PauliTerm to pyquil; accuracy test
    """
    pauli_term = OrqPauliTerm("2.25*Y0*X1*Z2*X4")

    pyquil_op = orq_to_pyquil(pauli_term)

    ground_truth = (
        PyquilPauliTerm("Y", 0)
        * PyquilPauliTerm("X", 1)
        * PyquilPauliTerm("Z", 2)
        * PyquilPauliTerm("X", 4)
        * 2.25
    )

    assert ground_truth == pyquil_op
    assert isinstance(pyquil_op, PyquilPauliTerm)


def test_orq_to_pyquil_zero():
    identity_term = OrqPauliTerm("I0", 0)
    forest_term = orq_to_pyquil(identity_term)
    ground_truth = PyquilPauliTerm("I", 0, 0)

    assert ground_truth == forest_term


def test_pyquil_pauliterm_to_orq():
    pyquil_term = PyquilPauliTerm("X", 0) * PyquilPauliTerm("Y", 5)
    orq_term = OrqPauliTerm("X0*Y5")

    test_orq_term = pyquil_to_orq(pyquil_term)
    assert test_orq_term == orq_term
    assert isinstance(test_orq_term, OrqPauliTerm)


def test_pyquil_paulisum_to_orq():
    pyquil_term = PyquilPauliSum([PyquilPauliTerm("X", 0) * PyquilPauliTerm("Y", 5)])
    orq_term = OrqPauliTerm("X0*Y5")

    test_orq_term = pyquil_to_orq(pyquil_term)
    assert test_orq_term == orq_term
    assert isinstance(test_orq_term, OrqPauliSum)


def test_pyquil_to_orq_raises_error_when_qubit_index_not_integer():
    with pytest.raises(ValueError) as e:
        pyquil_to_orq(PyquilPauliTerm("X", QubitPlaceholder()))
        assert "must be integer" in e.value
